@@ -80,6 +80,13 @@ bool aes_wrap(const uint8_t *kek, const uint8_t *in, size_t len, uint8_t *out);
 bool arc4_skip(const uint8_t *key, size_t key_len, size_t skip,
 		const uint8_t *in, size_t len, uint8_t *out);
 
+bool aes_siv_encrypt(const uint8_t *key, size_t key_len, const uint8_t *in,
+			size_t in_len, struct iovec *ad, size_t num_ad,
+			uint8_t *out);
+bool aes_siv_decrypt(const uint8_t *key, size_t key_len, const uint8_t *in,
+			size_t in_len, struct iovec *ad, size_t num_ad,
+			uint8_t *out);
+
 int crypto_cipher_key_len(enum crypto_cipher cipher);
 int crypto_cipher_tk_bits(enum crypto_cipher cipher);
 
@@ -107,7 +114,7 @@ bool crypto_derive_pairwise_ptk(const uint8_t *pmk, size_t pmk_len,
 				const uint8_t *addr1, const uint8_t *addr2,
 				const uint8_t *nonce1, const uint8_t *nonce2,
 				uint8_t *out_ptk, size_t ptk_len,
-				bool use_sha256);
+				enum l_checksum_type type);
 
 bool crypto_derive_pmk_r0(const uint8_t *xxkey,
 				const uint8_t *ssid, size_t ssid_len,
