@@ -25,21 +25,23 @@
 #endif
 
 #define _GNU_SOURCE
-#include <string.h>
-#include <stdarg.h>
 #include <stdio.h>
-#include <fcntl.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <stdlib.h>
-#include <unistd.h>
 #include <errno.h>
 #include <ctype.h>
+#include <fcntl.h>
+#include <string.h>
+#include <stdarg.h>
+#include <stdlib.h>
+#include <unistd.h>
+#include <time.h>
+#include <sys/time.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 
 #include <ell/ell.h>
 
-#include "common.h"
-#include "storage.h"
+#include "src/common.h"
+#include "src/storage.h"
 
 #ifdef TEMP_FAILURE_RETRY
 #define TFR TEMP_FAILURE_RETRY
@@ -183,12 +185,12 @@ char *storage_get_network_file_path(enum security type, const char *ssid)
 		hex = l_util_hexstring((const unsigned char *) ssid,
 					strlen(ssid));
 
-		path = l_strdup_printf(STORAGEDIR "/=%s.%s", hex,
+		path = l_strdup_printf(DAEMON_STORAGEDIR "/=%s.%s", hex,
 					security_to_str(type));
 
 		l_free(hex);
 	} else
-		path = l_strdup_printf(STORAGEDIR "/%s.%s", ssid,
+		path = l_strdup_printf(DAEMON_STORAGEDIR "/%s.%s", ssid,
 					security_to_str(type));
 
 	return path;
@@ -341,3 +343,33 @@ int storage_network_remove(enum security type, const char *ssid)
 
 	return ret < 0 ? -errno : 0;
 }
+
+static const char *known_freq_file_path =
+				DAEMON_STORAGEDIR "/.known_network.freq";
+
+struct l_settings *storage_known_frequencies_load(void)
+{
+	struct l_settings *known_freqs;
+
+	known_freqs = l_settings_new();
+
+	if (!l_settings_load_from_file(known_freqs, known_freq_file_path)) {
+		l_settings_free(known_freqs);
+		known_freqs = NULL;
+	}
+
+	return known_freqs;
+}
+
+void storage_known_frequencies_sync(struct l_settings *known_freqs)
+{
+	char *data;
+	size_t len;
+
+	if (!known_freqs)
+		return;
+
+	data = l_settings_to_data(known_freqs, &len);
+	write_file(data, len, "%s", known_freq_file_path);
+	l_free(data);
+}
