@@ -46,6 +46,8 @@
 #include <glob.h>
 #include <ell/ell.h>
 
+#include "linux/nl80211.h"
+
 #ifndef WAIT_ANY
 #define WAIT_ANY (-1)
 #endif
@@ -76,11 +78,14 @@ static bool valgrind;
 static char *gdb_opt;
 static bool enable_debug;
 const char *debug_filter;
+static struct l_settings *hw_config;
+static bool native_hw;
 static const char *qemu_binary;
 static const char *kernel_image;
 static const char *exec_home;
 static const char *test_action_params;
 static char top_level_path[PATH_MAX];
+struct l_queue *wiphy_list;
 
 #if defined(__i386__)
 /*
@@ -147,7 +152,7 @@ static const char * const qemu_table[] = { NULL };
 
 struct wiphy {
 	char name[20];
-	int id;
+	uint32_t id;
 	unsigned int interface_index;
 	bool interface_created : 1;
 	bool used_by_hostapd : 1;
@@ -293,7 +298,6 @@ static char *const qemu_argv[] = {
 	"",
 	"-machine", "type=q35,accel=kvm:tcg",
 	"-nodefaults",
-	"-nodefconfig",
 	"-no-user-config",
 	"-monitor", "none",
 	"-display", "none",
@@ -301,7 +305,6 @@ static char *const qemu_argv[] = {
 	"-nographic",
 	"-vga", "none",
 	"-net", "none",
-	"-balloon", "none",
 	"-no-acpi",
 	"-no-hpet",
 	"-no-reboot",
@@ -342,13 +345,16 @@ static bool check_virtualization(void)
 	return false;
 }
 
-static void start_qemu(void)
+static bool start_qemu(void)
 {
 	char cwd[PATH_MAX], testargs[PATH_MAX];
 	char *initcmd, *cmdline;
 	char **argv;
 	int i, pos;
 	bool has_virt;
+	int num_pci = 0, num_usb = 0;
+	char **pci_keys = NULL;
+	char **usb_keys = NULL;
 
 	has_virt = check_virtualization();
 
@@ -379,7 +385,7 @@ static void start_qemu(void)
 			"TESTVERBOUT=\'%s\' DEBUG_FILTER=\'%s\'"
 			"TEST_ACTION=%u TEST_ACTION_PARAMS=\'%s\' "
 			"TESTARGS=\'%s\' PATH=\'%s\' VALGRIND=%u"
-			"GDB=\'%s\'",
+			"GDB=\'%s\' HW=\'%s\'",
 			check_verbosity("kernel") ? "ignore_loglevel" : "quiet",
 			initcmd, cwd, verbose_opt ? verbose_opt : "none",
 			enable_debug ? debug_filter : "",
@@ -388,9 +394,40 @@ static void start_qemu(void)
 			testargs,
 			getenv("PATH"),
 			valgrind,
-			gdb_opt ? gdb_opt : "none");
+			gdb_opt ? gdb_opt : "none",
+			hw_config ? "real" : "virtual");
 
-	argv = alloca(sizeof(qemu_argv) + sizeof(char *) * 7);
+	if (hw_config) {
+		if (l_settings_has_group(hw_config, "PCIAdapters")) {
+			pci_keys = l_settings_get_keys(hw_config, "PCIAdapters");
+
+			for (num_pci = 0; pci_keys[num_pci]; num_pci++);
+		}
+
+		if (l_settings_has_group(hw_config, "USBAdapters")) {
+			usb_keys = l_settings_get_keys(hw_config, "USBAdapters");
+
+			for (num_usb = 0; usb_keys[num_usb]; num_usb++);
+		}
+
+		if (!pci_keys && !usb_keys) {
+			l_error("hs config had no PCIAdapters or USBAdapters");
+			l_free(initcmd);
+			l_free(cmdline);
+			return false;
+		}
+	}
+
+	/*
+	 * This got quite confusing. We need enough room for:
+	 *
+	 * qemu_argv (static list above with default parameters)
+	 * -kernel,-append,-cpu,-host parameters (7)
+	 * -enable-kvm and/or -usb (2)
+	 * PCI and/or USB parameters (num_pci * 2) (num_usb * 2)
+	 */
+	argv = alloca(sizeof(qemu_argv) + sizeof(char *) *
+				(7 + (2 + (num_pci * 2) + (num_usb * 2))));
 	memcpy(argv, qemu_argv, sizeof(qemu_argv));
 
 	pos = (sizeof(qemu_argv) / sizeof(char *)) - 1;
@@ -404,6 +441,41 @@ static void start_qemu(void)
 	argv[pos++] = "-cpu";
 	argv[pos++] = has_virt ? "host" : "max";
 
+	if (pci_keys) {
+		argv[pos++] = "-enable-kvm";
+		for (i = 0; pci_keys[i]; i++) {
+			argv[pos++] = "-device";
+			argv[pos] = alloca(22);
+			sprintf(argv[pos], "vfio-pci,host=%s",
+					l_settings_get_value(hw_config,
+					"PCIAdapters", pci_keys[i]));
+			pos++;
+		}
+	}
+
+	if (usb_keys) {
+		argv[pos++] = "-usb";
+		for (i = 0; usb_keys[i]; i++) {
+			const char *value = l_settings_get_value(hw_config,
+						"USBAdapters", usb_keys[i]);
+			char **info = l_strsplit(value, ',');
+
+			if (l_strv_length(info) != 2) {
+				l_error("hw config formatting error");
+				l_strv_free(info);
+				return false;
+			}
+
+			argv[pos++] = "-device";
+			argv[pos] = alloca(32);
+			sprintf(argv[pos], "usb-host,hostbus=%s,hostaddr=%s",
+						info[0], info[1]);
+			pos++;
+
+			l_strv_free(info);
+		}
+	}
+
 	argv[pos] = NULL;
 
 	execve(argv[0], argv, qemu_envp);
@@ -411,6 +483,8 @@ static void start_qemu(void)
 	/* Don't expect to reach here */
 	free(initcmd);
 	free(cmdline);
+
+	return true;
 }
 
 static pid_t execute_program(char *argv[], bool wait, bool verbose)
@@ -668,14 +742,14 @@ static bool list_hwsim_radios(void)
 	return true;
 }
 
-static int read_radio_id(void)
+static uint32_t read_radio_id(void)
 {
 	static int current_radio_id;
 
-	return current_radio_id++;
+	return ++current_radio_id;
 }
 
-static int create_hwsim_radio(const char *radio_name,
+static uint32_t create_hwsim_radio(const char *radio_name,
 				const unsigned int channels, bool p2p_device,
 							bool use_chanctx)
 {
@@ -692,7 +766,7 @@ static int create_hwsim_radio(const char *radio_name,
 
 	pid = execute_program(argv, true, check_verbosity(BIN_HWSIM));
 	if (pid < 0)
-		return -1;
+		return 0;
 
 	return read_radio_id();
 }
@@ -971,6 +1045,7 @@ static bool find_test_configuration(const char *path, int level,
 #define HW_CONFIG_SETUP_START_IWD	"start_iwd"
 #define HW_CONFIG_SETUP_IWD_CONF_DIR	"iwd_config_dir"
 #define HW_CONFIG_SETUP_REG_DOMAIN	"reg_domain"
+#define HW_CONFIG_SETUP_NEEDS_HWSIM	"needs_hwsim"
 
 static struct l_settings *read_hw_config(const char *test_dir_path)
 {
@@ -1015,7 +1090,7 @@ static bool configure_hw_radios(struct l_settings *hw_settings,
 						struct l_queue *wiphy_list)
 {
 	char **radio_conf_list;
-	int i, num_radios_requested, num_radios_created;
+	int num_radios_requested, num_radios_created;
 	bool status = false;
 	bool has_hw_conf;
 
@@ -1036,8 +1111,14 @@ static bool configure_hw_radios(struct l_settings *hw_settings,
 		l_settings_get_string_list(hw_settings, HW_CONFIG_GROUP_SETUP,
 						HW_CONFIG_SETUP_RADIO_CONFS,
 									':');
+	if (has_hw_conf && !radio_conf_list) {
+		l_error("%s doesn't parse", HW_CONFIG_SETUP_RADIO_CONFS);
+		return false;
+	}
 
 	if (has_hw_conf) {
+		int i;
+
 		for (i = 0; radio_conf_list[i]; i++) {
 			size_t len = strlen(radio_conf_list[i]);
 
@@ -1069,7 +1150,6 @@ static bool configure_hw_radios(struct l_settings *hw_settings,
 	}
 
 	num_radios_created = 0;
-	i = 0;
 
 	while (num_radios_requested > num_radios_created) {
 		struct wiphy *wiphy;
@@ -1085,13 +1165,11 @@ static bool configure_hw_radios(struct l_settings *hw_settings,
 			p2p_device = true;
 			use_chanctx = true;
 
-			has_hw_conf = false;
-
 			sprintf(wiphy->name, "rad%d", num_radios_created);
 			goto configure;
 		}
 
-		strcpy(wiphy->name, radio_conf_list[i]);
+		strcpy(wiphy->name, radio_conf_list[num_radios_created]);
 
 		if (!l_settings_get_uint(hw_settings, wiphy->name,
 					HW_CONFIG_PHY_CHANNELS, &channels))
@@ -1109,28 +1187,12 @@ configure:
 		wiphy->id = create_hwsim_radio(wiphy->name, channels,
 						p2p_device, use_chanctx);
 
-		if (wiphy->id < 0) {
+		if (wiphy->id == 0) {
 			l_free(wiphy);
 			goto exit;
 		}
 
-		l_queue_push_head(wiphy_list, wiphy);
-
-		wiphy->interface_name = l_strdup_printf("%s%d",
-							HW_INTERFACE_PREFIX,
-							num_radios_created);
-		if (!create_interface(wiphy->interface_name, wiphy->name))
-			goto exit;
-
-		wiphy->interface_created = true;
-		wiphy->interface_index = num_radios_created;
-		l_info("Created interface %s on %s radio",
-			wiphy->interface_name, wiphy->name);
-
-		if (!set_interface_state(wiphy->interface_name,
-						HW_INTERFACE_STATE_UP))
-			goto exit;
-
+		l_queue_push_tail(wiphy_list, wiphy);
 		num_radios_created++;
 	}
 
@@ -1154,15 +1216,20 @@ static void wiphy_free(void *data)
 		else
 			l_error("Failed to remove interface %s",
 				wiphy->interface_name);
-
-		l_free(wiphy->interface_name);
 	}
 
-	destroy_hwsim_radio(wiphy->id);
-	l_debug("Removed radio id %d", wiphy->id);
+	/* Native interfaces cannot be destroyed */
+	if (native_hw) {
+		set_interface_state(wiphy->interface_name,
+					HW_INTERFACE_STATE_DOWN);
+	} else {
+		destroy_hwsim_radio(wiphy->id);
+		l_debug("Removed radio id %d", wiphy->id);
+	}
 
 	l_free(wiphy->hostapd_config);
 	l_free(wiphy->hostapd_ctrl_interface);
+	l_free(wiphy->interface_name);
 
 	l_free(wiphy);
 }
@@ -1170,13 +1237,16 @@ static void wiphy_free(void *data)
 static bool configure_hostapd_instances(struct l_settings *hw_settings,
 						char *config_dir_path,
 						struct l_queue *wiphy_list,
-						pid_t hostapd_pids_out[])
+						pid_t hostapd_pids_out[],
+						int *phys_used)
 {
 	char **hostap_keys;
 	int i;
 	char **hostapd_config_file_paths;
 	struct wiphy **wiphys;
 
+	*phys_used = 0;
+
 	if (!l_settings_has_group(hw_settings, HW_CONFIG_GROUP_HOSTAPD)) {
 		l_info("No hostapd instances to create");
 		return true;
@@ -1196,6 +1266,7 @@ static bool configure_hostapd_instances(struct l_settings *hw_settings,
 	for (i = 0; hostap_keys[i]; i++) {
 		const struct l_queue_entry *wiphy_entry;
 		const char *hostapd_config_file;
+		unsigned wiphy_idx = 0;
 
 		hostapd_config_file =
 			l_settings_get_value(hw_settings,
@@ -1215,13 +1286,28 @@ static bool configure_hostapd_instances(struct l_settings *hw_settings,
 
 		for (wiphy_entry = l_queue_get_entries(wiphy_list);
 					wiphy_entry;
-					wiphy_entry = wiphy_entry->next) {
+					wiphy_entry = wiphy_entry->next,
+					wiphy_idx++) {
 			struct wiphy *wiphy = wiphy_entry->data;
 
-			if (strcmp(wiphy->name, hostap_keys[i]))
+			/*
+			 * We can skip this check in native mode since we have
+			 * no control over the phy name. Any test requiring a
+			 * "special" radio should not be ran in native mode.
+			 */
+			if (!native_hw && strcmp(wiphy->name, hostap_keys[i]))
 				continue;
 
 			if (wiphy->used_by_hostapd) {
+				/*
+				 * Since we bypass the above check in native
+				 * mode we could still get here. We can just
+				 * continue searching for more adapters if this
+				 * one is already in use.
+				 */
+				if (native_hw)
+					continue;
+
 				l_error("Wiphy %s already used by hostapd",
 					wiphy->name);
 				goto done;
@@ -1232,7 +1318,29 @@ static bool configure_hostapd_instances(struct l_settings *hw_settings,
 		}
 
 		if (!wiphy_entry) {
-			l_error("Failed to find available interface.");
+			l_error("Failed to find available wiphy.");
+			goto done;
+		}
+
+		wiphys[i]->interface_name = l_strdup_printf("%s%d",
+							HW_INTERFACE_PREFIX,
+							wiphy_idx);
+		if (!create_interface(wiphys[i]->interface_name,
+					wiphys[i]->name)) {
+			l_error("Failed to create hostapd interface %s on "
+				"radio %s",
+				wiphys[i]->interface_name, wiphys[i]->name);
+			goto done;
+		}
+
+		wiphys[i]->interface_created = true;
+		l_info("Created hostapd interface %s on %s radio",
+			wiphys[i]->interface_name, wiphys[i]->name);
+
+		if (!native_hw && !set_interface_state(wiphys[i]->interface_name,
+						HW_INTERFACE_STATE_UP)) {
+			l_error("Failed to set %s state UP",
+				wiphys[i]->interface_name);
 			goto done;
 		}
 
@@ -1241,6 +1349,8 @@ static bool configure_hostapd_instances(struct l_settings *hw_settings,
 			l_strdup_printf("%s/%s", HOSTAPD_CTRL_INTERFACE_PREFIX,
 					wiphys[0]->interface_name);
 		wiphys[i]->hostapd_config = l_strdup(hostapd_config_file);
+
+		(*phys_used)++;
 	}
 
 	hostapd_pids_out[0] = start_hostapd(hostapd_config_file_paths, wiphys);
@@ -1256,7 +1366,7 @@ done:
 }
 
 static pid_t start_iwd(const char *config_dir, struct l_queue *wiphy_list,
-		const char *ext_options)
+		const char *ext_options, int num_phys)
 {
 	char *argv[13];
 	char *iwd_phys = NULL;
@@ -1292,13 +1402,19 @@ static pid_t start_iwd(const char *config_dir, struct l_queue *wiphy_list,
 					wiphy_entry = wiphy_entry->next) {
 			struct wiphy *wiphy = wiphy_entry->data;
 
-			if (!wiphy->interface_created)
-				continue;
-
 			if (wiphy->used_by_hostapd)
 				continue;
 
+			/*
+			 * Break out, only adding the required number of phys
+			 * for this test.
+			 */
+			if (num_phys == 0)
+				break;
+
 			l_string_append_printf(list, "%s,", wiphy->name);
+
+			num_phys--;
 		}
 
 		iwd_phys = l_string_unwrap(list);
@@ -1664,10 +1780,9 @@ static void set_wiphy_list(struct l_queue *wiphy_list)
 		struct wiphy *wiphy = wiphy_entry->data;
 
 		size += 32 + strlen(wiphy->name);
-		if (wiphy->interface_created)
-			size += 32 + strlen(wiphy->interface_name);
 		if (wiphy->used_by_hostapd) {
-			size += 32 + strlen(wiphy->hostapd_ctrl_interface) +
+			size += 32 + strlen(wiphy->interface_name) +
+				strlen(wiphy->hostapd_ctrl_interface) +
 				strlen(wiphy->hostapd_config);
 		}
 	}
@@ -1682,12 +1797,13 @@ static void set_wiphy_list(struct l_queue *wiphy_list)
 		if (size)
 			var[size++] = '\n';
 
-		size += sprintf(var + size, "%s=%s=", wiphy->name,
-				wiphy->interface_name);
+		size += sprintf(var + size, "%s=", wiphy->name);
 
 		if (wiphy->used_by_hostapd)
 			size += sprintf(var + size,
-					"hostapd,ctrl_interface=%s,config=%s",
+					"hostapd,name=%s,ctrl_interface=%s,"
+					"config=%s",
+					wiphy->interface_name,
 					wiphy->hostapd_ctrl_interface,
 					wiphy->hostapd_config);
 		else
@@ -1712,6 +1828,25 @@ static void set_reg_domain(const char *domain)
 	execute_program(argv, false, false);
 }
 
+static void wiphy_up(void *data, void *user_data)
+{
+	struct wiphy *wiphy = data;
+
+	set_interface_state(wiphy->interface_name, true);
+}
+
+static void wiphy_reset(void *data, void *user_data)
+{
+	struct wiphy *wiphy = data;
+
+	wiphy->used_by_hostapd = false;
+
+	l_free(wiphy->hostapd_config);
+	wiphy->hostapd_config = NULL;
+	l_free(wiphy->hostapd_ctrl_interface);
+	wiphy->hostapd_ctrl_interface = NULL;
+}
+
 static void create_network_and_run_tests(const void *key, void *value,
 								void *data)
 {
@@ -1724,14 +1859,16 @@ static void create_network_and_run_tests(const void *key, void *value,
 	char *iwd_config_dir;
 	char **tmpfs_extra_stuff = NULL;
 	struct l_settings *hw_settings;
-	struct l_queue *wiphy_list;
 	struct l_queue *test_queue;
 	struct l_queue *test_stats_queue;
 	bool start_iwd_daemon = true;
+	bool needs_hwsim = false;
 	bool ofono_req = false;
 	const char *sim_keys;
 	const char *iwd_ext_options = NULL;
 	const char *reg_domain;
+	int phys_used;
+	int num_radios;
 
 	memset(hostapd_pids, -1, sizeof(hostapd_pids));
 
@@ -1751,7 +1888,6 @@ static void create_network_and_run_tests(const void *key, void *value,
 	if (!hw_settings)
 		return;
 
-	wiphy_list = l_queue_new();
 	l_info("Configuring network...");
 
 	if (chdir(config_dir_path) < 0) {
@@ -1801,34 +1937,75 @@ static void create_network_and_run_tests(const void *key, void *value,
 	if (!create_tmpfs_extra_stuff(tmpfs_extra_stuff))
 		goto remove_abs_paths;
 
-	reg_domain = l_settings_get_value(hw_settings, HW_CONFIG_GROUP_SETUP,
+	l_settings_get_int(hw_settings, HW_CONFIG_GROUP_SETUP,
+					HW_CONFIG_SETUP_NUM_RADIOS,
+					&num_radios);
+
+	if (!native_hw) {
+		reg_domain = l_settings_get_value(hw_settings,
+						HW_CONFIG_GROUP_SETUP,
 						HW_CONFIG_SETUP_REG_DOMAIN);
-	if (reg_domain)
-		set_reg_domain(reg_domain);
+		if (reg_domain)
+			set_reg_domain(reg_domain);
 
-	if (!configure_hw_radios(hw_settings, wiphy_list))
-		goto remove_abs_paths;
+		wiphy_list = l_queue_new();
 
-	medium_pid = register_hwsim_as_trans_medium();
-	if (medium_pid < 0)
-		goto remove_abs_paths;
+		if (!configure_hw_radios(hw_settings, wiphy_list))
+			goto remove_abs_paths;
 
-	if (check_verbosity("hwsim")) {
-		list_hwsim_radios();
-		list_interfaces();
+		medium_pid = register_hwsim_as_trans_medium();
+		if (medium_pid < 0)
+			goto remove_abs_paths;
+
+		if (check_verbosity("hwsim")) {
+			list_hwsim_radios();
+			list_interfaces();
+		}
+	} else {
+		int len;
+
+		l_settings_get_bool(hw_settings, HW_CONFIG_GROUP_SETUP,
+				HW_CONFIG_SETUP_NEEDS_HWSIM, &needs_hwsim);
+
+		/* Skip test that require hwsim dbus APIs (hwsim not running) */
+		if (needs_hwsim) {
+			l_error("test requires hwsim, skipping");
+			goto remove_abs_paths;
+		}
+
+		len = l_queue_length(wiphy_list);
+
+		/* Skip tests that need more radios than we have */
+		if (num_radios > len) {
+			l_error("test requires %d radios, only %d found",
+					num_radios, len);
+			goto remove_abs_paths;
+		}
+
+		l_queue_foreach(wiphy_list, wiphy_up, NULL);
 	}
 
 	if (check_verbosity("tls"))
 		setenv("IWD_TLS_DEBUG", "on", true);
 
 	if (!configure_hostapd_instances(hw_settings, config_dir_path,
-						wiphy_list, hostapd_pids))
+						wiphy_list, hostapd_pids,
+						&phys_used))
 		goto exit_hostapd;
 
 	l_settings_get_bool(hw_settings, HW_CONFIG_GROUP_SETUP,
 				HW_CONFIG_SETUP_START_IWD, &start_iwd_daemon);
 
 	if (start_iwd_daemon) {
+		/*
+		 * In native mode we may have more radios than a test actually
+		 * needs. This would result in IWD managing all phys that
+		 * hostapd wasn't using, which could throw off test results.
+		 * By passing the number of phys the test expects IWD to have
+		 * we can leave the remaining (unneeded) phys unmanaged.
+		 */
+		int iwd_phys = num_radios - phys_used;
+
 		iwd_config_dir =
 			l_settings_get_string(hw_settings,
 						HW_CONFIG_GROUP_SETUP,
@@ -1837,7 +2014,7 @@ static void create_network_and_run_tests(const void *key, void *value,
 			iwd_config_dir = DAEMON_CONFIGDIR;
 
 		iwd_pid = start_iwd(iwd_config_dir, wiphy_list,
-				iwd_ext_options);
+				iwd_ext_options, iwd_phys);
 
 		if (iwd_pid == -1)
 			goto exit_hostapd;
@@ -1875,13 +2052,23 @@ static void create_network_and_run_tests(const void *key, void *value,
 exit_hostapd:
 	destroy_hostapd_instances(hostapd_pids);
 
-	terminate_medium(medium_pid);
+	if (!native_hw)
+		terminate_medium(medium_pid);
 
 remove_abs_paths:
 	remove_absolute_path_dirs(tmpfs_extra_stuff);
 
 exit_hwsim:
-	l_queue_destroy(wiphy_list, wiphy_free);
+	/*
+	 * If running in hwsim mode, we want to completely free/destroy the
+	 * wiphy list since it will be re-populated on the next test. For the
+	 * native case we want to reset the list as if it was freshly
+	 * discovered. This ensures that all the hostapd flags get reset.
+	 */
+	if (!native_hw)
+		l_queue_destroy(wiphy_list, wiphy_free);
+	else
+		l_queue_foreach(wiphy_list, wiphy_reset, NULL);
 
 	l_settings_free(hw_settings);
 	l_strfreev(tmpfs_extra_stuff);
@@ -2111,6 +2298,200 @@ exit:
 	l_strfreev(unit_tests);
 }
 
+static bool wiphy_match(const void *a, const void *b)
+{
+	const struct wiphy *wiphy = a;
+	uint32_t id = L_PTR_TO_UINT(b);
+
+	return (wiphy->id == id);
+}
+
+static struct wiphy *wiphy_find(int wiphy_id)
+{
+	return l_queue_find(wiphy_list, wiphy_match, L_UINT_TO_PTR(wiphy_id));
+}
+
+static void wiphy_dump_callback(struct l_genl_msg *msg, void *user_data)
+{
+	struct wiphy *wiphy;
+	struct l_genl_attr attr;
+	uint32_t id;
+	uint16_t type, len;
+	const void *data;
+	const char *name;
+	uint32_t name_len;
+
+	if (!l_genl_attr_init(&attr, msg))
+		return;
+
+	/*
+	 * The wiphy attribute, name and generation are always the first
+	 * three attributes (in that order) in every NEW_WIPHY & DEL_WIPHY
+	 * message.  If not, then error out with a warning and ignore the
+	 * whole message.
+	 */
+	if (!l_genl_attr_next(&attr, &type, &len, &data))
+		return;
+
+	if (type != NL80211_ATTR_WIPHY)
+		return;
+
+	if (len != sizeof(uint32_t))
+		return;
+
+	id = *((uint32_t *) data);
+
+	if (wiphy_find(id))
+		return;
+
+	if (!l_genl_attr_next(&attr, &type, &len, &data))
+		return;
+
+	if (type != NL80211_ATTR_WIPHY_NAME)
+		return;
+
+	if (len > sizeof(((struct wiphy *) 0)->name))
+		return;
+
+	name = data;
+	name_len = len;
+
+	wiphy = l_new(struct wiphy, 1);
+	strncpy(wiphy->name, name, name_len);
+	wiphy->id = id;
+
+	l_queue_push_tail(wiphy_list, wiphy);
+}
+
+static void iface_dump_callback(struct l_genl_msg *msg, void *user_data)
+{
+	struct l_genl_attr attr;
+	uint16_t type, len;
+	const void *data;
+	const char *ifname = NULL;
+	struct wiphy *wiphy = NULL;
+
+	if (!l_genl_attr_init(&attr, msg))
+		return;
+
+	while (l_genl_attr_next(&attr, &type, &len, &data)) {
+		switch (type) {
+
+		case NL80211_ATTR_IFNAME:
+			if (len > 16) {
+				l_warn("Invalid interface name attribute");
+				return;
+			}
+
+			ifname = data;
+			break;
+
+		case NL80211_ATTR_WIPHY:
+			if (len != sizeof(uint32_t)) {
+				l_warn("Invalid wiphy attribute");
+				return;
+			}
+
+			wiphy = wiphy_find(*((uint32_t *) data));
+			break;
+		}
+	}
+
+	if (!ifname || !wiphy)
+		return;
+
+	wiphy->interface_name = l_strdup(ifname);
+	wiphy->interface_created = false;
+
+	l_info("Discovered interface %s", wiphy->interface_name);
+}
+
+struct nl_data {
+	struct l_genl *genl;
+	struct l_genl_family *nl80211;
+};
+
+static void iface_dump_done(void *user_data)
+{
+	struct nl_data *data = user_data;
+
+	l_debug("Interface discovery complete, running tests");
+
+	list_interfaces();
+
+	run_auto_tests();
+
+	l_queue_destroy(wiphy_list, wiphy_free);
+
+	l_genl_family_unref(data->nl80211);
+	l_genl_unref(data->genl);
+	l_free(data);
+
+	l_main_quit();
+}
+
+static void wiphy_dump_done(void *user_data)
+{
+	struct nl_data *data = user_data;
+	struct l_genl_msg *msg;
+
+	l_debug("Wiphy discovery complete, discovering interfaces");
+
+	msg = l_genl_msg_new(NL80211_CMD_GET_INTERFACE);
+	if (!l_genl_family_dump(data->nl80211, msg, iface_dump_callback,
+						data, iface_dump_done))
+		l_error("Getting all interface information failed");
+}
+
+static void nl80211_appeared(void *user_data)
+{
+	struct nl_data *data = user_data;
+	struct l_genl_msg *msg;
+
+	wiphy_list = l_queue_new();
+
+	l_debug("Found nl80211 interface");
+
+	msg = l_genl_msg_new(NL80211_CMD_GET_WIPHY);
+	if (!l_genl_family_dump(data->nl80211, msg, wiphy_dump_callback,
+						data, wiphy_dump_done))
+		l_error("Getting all wiphy devices failed");
+}
+
+static void nl80211_vanished(void *user_data)
+{
+	l_debug("Lost nl80211 interface");
+
+	l_main_quit();
+}
+
+static void start_hw_discovery(void)
+{
+	struct nl_data *data = l_new(struct nl_data, 1);
+
+	data->genl = l_genl_new_default();
+	data->nl80211 = l_genl_family_new(data->genl, NL80211_GENL_NAME);
+
+	l_genl_family_set_watches(data->nl80211, nl80211_appeared,
+					nl80211_vanished, data, NULL);
+
+	/*
+	 * This is somewhat of a mystery, but it appears that
+	 * calling lshw causes the OS to re-enumerate the USB
+	 * bus. Without this no USB adapters are found when
+	 * doing the wiphy/iface dump from nl80211.
+	 *
+	 * This also conveniently prints all the network
+	 * adapters and their iface name, so its much easier
+	 * to know which adapter are being used by iwd/hostapd
+	 * after the test.
+	 */
+	if (system("lshw -C network"))
+		l_info("lshw failed");
+
+	l_main_run();
+}
+
 static void run_tests(void)
 {
 	char cmdline[CMDLINE_MAX], *ptr, *cmds;
@@ -2133,6 +2514,20 @@ static void run_tests(void)
 		return;
 	}
 
+	ptr = strstr(cmdline, "HW=");
+	if (ptr) {
+		*ptr = '\0';
+		test_action_str = ptr + 4;
+
+		ptr = strchr(test_action_str, '\'');
+		*ptr = '\0';
+
+		if (!strcmp(test_action_str, "virtual"))
+			native_hw = false;
+		else
+			native_hw = true;
+	}
+
 	ptr = strstr(cmdline, "GDB=");
 	if (ptr) {
 		*ptr = '\0';
@@ -2280,7 +2675,10 @@ static void run_tests(void)
 
 	switch (test_action) {
 	case ACTION_AUTO_TEST:
-		run_auto_tests();
+		if (native_hw)
+			start_hw_discovery();
+		else
+			run_auto_tests();
 		break;
 	case ACTION_UNIT_TEST:
 		run_unit_tests();
@@ -2305,7 +2703,9 @@ static void usage(void)
 						"\t\t\t\tto see valgrind"
 						" output\n"
 		"\t-g, --gdb <iwd|hostapd>	Run gdb on the specified"
-						" executable");
+						" executable\n"
+		"\t-w, --hw <config>	Run using a physical hardware "
+					"configuration");
 	l_info("Commands:\n"
 		"\t-A, --auto-tests <dirs>	Comma separated list of the "
 						"test configuration\n\t\t\t\t"
@@ -2323,6 +2723,7 @@ static const struct option main_options[] = {
 	{ "debug",	optional_argument, NULL, 'd' },
 	{ "gdb",	required_argument, NULL, 'g' },
 	{ "valgrind",	no_argument,       NULL, 'V' },
+	{ "hw",		required_argument, NULL, 'w' },
 	{ "help",	no_argument,       NULL, 'h' },
 	{ }
 };
@@ -2334,6 +2735,9 @@ int main(int argc, char *argv[])
 	l_log_set_stderr();
 
 	if (getpid() == 1 && getppid() == 0) {
+		if (!l_main_init())
+			return EXIT_FAILURE;
+
 		prepare_sandbox();
 
 		run_tests();
@@ -2397,6 +2801,15 @@ int main(int argc, char *argv[])
 				return EXIT_FAILURE;
 			}
 			break;
+		case 'w':
+			hw_config = l_settings_new();
+			if (!l_settings_load_from_file(hw_config, optarg)) {
+				l_error("could not read hw config from %s",
+						optarg);
+				l_settings_free(hw_config);
+				return EXIT_FAILURE;
+			}
+			break;
 		case 'h':
 			usage();
 			return EXIT_SUCCESS;
@@ -2441,7 +2854,8 @@ int main(int argc, char *argv[])
 	l_info("Using QEMU binary %s", qemu_binary);
 	l_info("Using kernel image %s", kernel_image);
 
-	start_qemu();
+	if (!start_qemu())
+		return EXIT_FAILURE;
 
 	return EXIT_SUCCESS;
 }
